'''
    EEGNet-8,2
    https://iopscience.iop.org/article/10.1088/1741-2552/aace8c/meta
'''


import torch
from torch import Tensor, nn
from torchinfo import summary


class Conv2dWithNorm(nn.Conv2d):
    def __init__(self, *args, doWeightNorm=True, maxNorm=1, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.maxNorm = maxNorm
        self.doWeightNorm = doWeightNorm
        
    def forward(self, input: Tensor) -> Tensor:
        if self.doWeightNorm:
            self.weight.data = torch.renorm(
                self.weight.data, 2, 0, self.maxNorm
            )
        return super().forward(input)


class LinearWithNorm(nn.Linear):
    def __init__(self, *args, maxNorm, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.maxNorm = maxNorm

    def forward(self, input: Tensor) -> Tensor:
        self.weight.data = torch.renorm(
            self.weight.data, 2, 0, self.maxNorm
        )
        return super().forward(input)


class EEGNet(nn.Module):
    def __init__(self, nCh, nTime, C1=63, F1=8, D=2, F2=16, C2=15, P1=8, P2=16,
                 p=0.5, cls=4) -> None:
        super().__init__()

        self.filter = nn.Sequential(
            nn.Conv2d(1, F1, (1, C1), padding=(0, C1//2), bias=False),
            nn.BatchNorm2d(F1)
        )

        self.depthwiseConv = nn.Sequential(
            Conv2dWithNorm(F1, D*F1, (nCh, 1), groups=F1, bias=False, maxNorm=1,
                           doWeightNorm=False),
            nn.BatchNorm2d(D*F1),
            nn.ELU(),
            nn.AvgPool2d((1, P1)),
            nn.Dropout(p)
        )

        self.separableConv = nn.Sequential(
            nn.Conv2d(D*F1, D*F1, (1, C2), padding=(0, C2//2), groups=D*F1, bias=False),
            nn.Conv2d(D*F1, F2, 1, bias=False),
            nn.BatchNorm2d(F2),
            nn.ELU(),
            nn.AvgPool2d((1, P2)),
            nn.Dropout(p)
        )

        self.flatten = nn.Flatten()
        self.fc = nn.Sequential(
            nn.Linear(self.get_size(nCh, nTime), cls, bias=True),
            nn.LogSoftmax(dim=1)
        )

    def get_size(self, nCh, nTime):
        x = torch.randn(1, 1, nCh, nTime)
        out = self.filter(x)
        out = self.depthwiseConv(out)
        out = self.separableConv(out)
        return self.flatten(out).size(1)

    def forward(self, x):
        out = self.filter(x)
        out = self.depthwiseConv(out)
        out = self.separableConv(out)
        out = self.flatten(out)
        return self.fc(out)


if __name__ == '__main__':
    net = EEGNet(22, 1000).cuda()
    summary(net, (1, 1, 22, 1000))


#!/usr/bin/env python
# coding: utf-8

"""
    This module defines with what experiments the module is trained and validated.

    experiments:
        1 - kFoldCV : K-Fold cross-validation.
        2 - HoldOut : hold out validation.

    @Author  : SheepTAO
    @Time    : 2023-07-26
"""


import os, pickle
from torch import Tensor
from typing import Union, Optional, Tuple
from sklearn.model_selection import StratifiedKFold
from torchmetrics.functional.classification.cohen_kappa import cohen_kappa
from torchmetrics.aggregation import MeanMetric, CatMetric

from .train import Train
from ..data.datasets import EEGDataset
from ..tools import Timer, Filer, Logger
from ..utils import DPEEG_SEED, DPEEG_DIR
from .evaluate import save_cm_img


class KFoldCV:
    def __init__(
        self,
        trainer : Train,
        k : int = 10,
        outFolder : Optional[str] = None,
        maxEpochs_1 : int = 1500,
        maxEpochs_2 : int = 600,
        noIncreaseEpochs : int = 100,
        varCheck : str = 'valInacc',
        shuffle : bool = True,
        seed : int = DPEEG_SEED,
        verbose : Union[int, str] = 'INFO',
    ) -> None:
        '''K-Fold cross-validation class.

        Parameters
        ----------
        trainer : Train
            Trainer used for training module on dataset.
        k : int, optional
            k of k-Fold. Default is 10.
        outFolder : str, optional
            Store all experimental results in a folder named with the model 
            class name in the specified folder. Default is 
            '~/dpeeg/out/model_name/'.
        maxEpochs_1, maxEpochs_2 : int
            Maximum number of epochs in the x stage of training. Default is
            1500 and 600 respectively.
        noIncreaseEpochs : int
            Maximum number of consecutive epochs when the accuracy of the first
            stage validation set has no relative improvement. Default is 100.
        varCheck : str
            The best value ('valInacc'/'valLoss') to check while determining 
            the best model which will be used for parameter initialization in
            the second stage of model training. Default is 'valInacc'.
        shuffle : bool
            Wheter shuffle dataset. Default is True.
        seed : int
            Seed of random for review. Default is DPEEG_SEED.
        verbose : int, str
            The log level of console. Default is INFO. Mainly used for debugging.

        All experimental results of the model will be saved in the outFolder
        directory of the trainer parameter.
        TODO: Nested Cross-Validation.
        '''
        self.trainer = trainer
        self.k = k
        self.maxEpochs_1 = maxEpochs_1
        self.maxEpochs_2 = maxEpochs_2
        self.noIncreaseEpochs = noIncreaseEpochs
        self.varCheck = varCheck

        self.loger = Logger('dpeeg_exp', clevel=verbose)
        self.timer = Timer()

        # set output folder
        netName = trainer.net.__class__.__name__
        self.outFolder = os.path.join(os.path.abspath(outFolder), netName) \
            if outFolder else os.path.join(DPEEG_DIR, 'out', netName)
        os.makedirs(self.outFolder, exist_ok=True)
        self.loger.info(f'Results will be saved in folder: {self.outFolder}')

        # create a stratified k-fold index
        self.skf = StratifiedKFold(k, shuffle=shuffle, random_state=seed)

    def kfold(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        clsName : Union[tuple, list],
        sub : Optional[str] = None,
    ) -> Tuple[Tensor, Tensor, dict]:
        '''Basic K-Fold cross-validation function.

        Parameters
        ----------
        trainset: tuple, list
            Dataset used for training module. If type is tuple or list, dataset
            should be (data, labels).
        testset:
            Dataset used for evaluating moduel. If type is tuple or list, dataset
            should be (data, labels).
        sub : str, optional
            Create a subdirectory of outFolder to store all yield results during 
            KFold_CV. If None, will use outFolder. Default is None.
        clsName : tuple, list
            The name of dataset labels.
        
        Returns
        -------
        Return testAcc, testKappa and results :
        
        results = {
            'expNo_x' : { ... },
            .
            .
            .
        }
        '''
        # save all exp results
        results = {}
        trainAccMetric, valAccMetric, testAccMetric = \
            MeanMetric(), MeanMetric(), MeanMetric()
        testPredsMetric, testTargetMetric = CatMetric(), CatMetric()

        # set kFold-cv results storage path
        subFolder = os.path.join(self.outFolder, sub) if sub else self.outFolder
        os.makedirs(subFolder, exist_ok=True)

        # store experiment results
        filer = Filer(os.path.join(subFolder, 'summary.txt'))
        cmFolder = os.path.join(subFolder, 'confusion_matrix')
        os.makedirs(cmFolder)

        # register a timer
        self.timer.start('kfold')

        # run k-fold
        for idx, (trainIdx, valIdx) in enumerate(self.skf.split(*trainset)):
            self.loger.info(f'\n# ------ {sub} - expNo.{idx+1} ------ #')

            # train one fold
            trainData = (trainset[0][trainIdx], trainset[1][trainIdx])
            valData = (trainset[0][valIdx], trainset[1][valIdx])
            expPath = os.path.join(subFolder, f'expNo_{idx + 1}')
            res = self.trainer.fit_with_val(
                trainData, valData, testset, expPath, clsName, self.maxEpochs_1,
                self.maxEpochs_2, self.noIncreaseEpochs, self.varCheck
            )

            # save all results
            results[f'expNo_{idx}'] = res
            trainRes, valRes, testRes = res['train'], res['val'], res['test']
            trainAccMetric.update(trainRes['acc'])
            valAccMetric.update(valRes['acc'])
            testAccMetric.update(testRes['acc'])
            testPredsMetric.update(testRes['preds'])
            testTargetMetric.update(testRes['target'])

            # save the confusion matirx image
            save_cm_img(testRes['preds'], testRes['target'], clsName,
                        os.path.join(cmFolder, f'expNo_{idx+1}_CM.png')
            )

            filer.write(f'------------------ expNo_{idx+1} ------------------\n')
            filer.write(f'Acc : Train={trainRes["acc"]:.4f} | Val={valRes["acc"]:.4f}' +
                        f' | Test={testRes["acc"]:.4f}\n\n')

        h, m, s = self.timer.stop('kfold')
        self.loger.info(f'\n[Cross-Validation Finish]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')

        # calculate the average acc
        trainAcc = trainAccMetric.compute()
        valAcc = valAccMetric.compute()
        testAcc = testAccMetric.compute()
        self.loger.info(f'Acc : train={trainAcc:.4f} | val={valAcc:.4f} | test={testAcc:.4f}') 
        
        # calculate cohen kappa
        testKappa = cohen_kappa(
            testPredsMetric.compute(), testTargetMetric.compute(),
            task = 'multiclass', num_classes = len(clsName)
        )

        # store the subject accuracy and cohen kappa on test set
        filer.write(f'--------- TestAvg ---------\n')
        filer.write(f'Acc = {testAcc*100:.2f}% | Kappa = {testKappa:.2f}\n')

        return testAcc, testKappa, results

    def run(
        self, 
        dataset : Union[EEGDataset, dict],
        clsName : Union[tuple, list], 
        datasetName : Optional[str] = None,
        desc : Optional[str] = None,
    ) -> dict:
        '''Run K-Fold cross validation on eeg datasets.

        Will select train set to process k-fold cross validation `kFold`, and 
        test set to evaluate the model.

        Parameters
        ----------
        dataset : EEGDataset, dict
            Dataset used for k-fold. Dataset structured as:
            
            {0 : { 'train': (array(nsample x channels x times), array(nsample)),
                   'test': (array(nsample x channels x times), array(nsample))},
             1 : {...},
            ...}
            
            where array can be `numpy.ndarray` or `torch.Tensor`. Please make 
            sure that the dataset has been split.
        clsName : tuple, list
            The name of dataset labels.
        datasetName : str, optional
            The dataset name to use. If None, `dataset.__class__.__name__` will 
            be used as the folder to save experimental results. If the type of 
            the dataset is dict, Please provide a dataset name to prevent 
            confusing results.
        desc : str, optional
            Add a short description to the current experiment. Default is None.

        Returns
        -------
        Return a dict of all subjects and corresponding experimental results.
        '''
        self.loger.info(f'\n[{self.k}-Fold Cross-Validation Starting]')

        # store all subjects' results in one folder
        self.outFolder = os.path.join(self.outFolder, dataset.__class__.__name__) \
            if not datasetName else os.path.join(self.outFolder, datasetName)
        os.makedirs(self.outFolder)
        filer = Filer(os.path.join(self.outFolder, 'summary.txt'))
        filer.write(f'[Start Time]: {self.timer.ctime()}\n\n')
        filer.write(f'[ExP Description]: {desc}\n\n')
        filer.write(str(self.trainer) + '\n\n')

        # save all sub results
        results = {}
        testAccMetric, testKappaMetric = MeanMetric(), MeanMetric()

        # update root timer and start k-fold cross validation for each subject
        self.timer.start()
        self.loger.info('=' * 50)
        for sub, data in dataset.items():

            self.loger.info(f'\n[Subject-{sub} Training ...]')
            self.loger.info(f'Train set = {data["train"][1].shape} | Test set = '
                  f'{data["test"][1].shape}')
            self.loger.info('-' * 50)

            # run a k-Fold cross validation
            testAcc, testKappa, subExpRes = \
                self.kfold(data['train'], data['test'], clsName, f'sub{sub}')

            results[f'sub_{sub}'] = subExpRes
            testAccMetric.update(testAcc)
            testKappaMetric.update(testKappa)

            filer.write(f'---------- Sub_{sub} ----------\n')
            filer.write(f'Acc = {testAcc*100:.2f}% | Kappa = {testKappa:.2f}\n\n')

        acc = testAccMetric.compute()
        kappa = testKappaMetric.compute()

        # store the model accuracy and cohen kappa
        filer.write(f'---------- MODEL ----------\n')
        filer.write(f'Acc = {acc*100:.2f}% | Kappa = {kappa:.2f}\n')
        with open(os.path.join(self.outFolder, f'results.pkl'), 'wb') as f:
            pickle.dump(results, f)

        # end
        h, m, s = self.timer.stop()
        self.loger.info(f'\n[All subjects finished]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')
        self.loger.info('=' * 50)
        self.loger.info(f'[Acc = {acc*100:.2f}% | Kappa = {kappa:.2f}]')

        return results


class HoldOut:
    '''TODO'''
    def __init__(self) -> None:
        pass



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to evaluate the metrics of the model and save.

    @author: SheepTAO
    @data: 2023-4-13
"""


import torch
import seaborn as sns
import matplotlib.pyplot as plt
from torch import Tensor
from typing import Union, List, Optional
from torchmetrics.functional.classification.confusion_matrix import confusion_matrix


def save_cm_img(
    preds : Tensor,
    target : Tensor,
    clsName : Union[list, tuple],
    figPath : str,
) -> None:
    '''Calculate and save the corresponding confusion matrix figure to the given
    img path.

    Parameters
    ----------
    preds : Tensor
        Predicted labels, as returned by a classifier.
    target : Tensor
        Ground truth (correct) labels.
    clsName : list, tuple
        The name of dataset labels.
    figPath : str
        Path to store the figure.
    '''
    cm = confusion_matrix(preds, target, 'multiclass', num_classes=len(clsName))
    fig = sns.heatmap(cm / cm.sum(1, keepdim=True), annot=True, cmap='Blues',
                      xticklabels=clsName, yticklabels=clsName) # type: ignore
    plt.savefig(figPath)
    # !!! Clear the current figure
    plt.clf()

def ttest_corrected(
    score1 : Tensor,
    score2 : Tensor,
    n1 : int,
    n2 : int,
) -> None:
    '''Corrected Paired t test for comparing the performance of two models.

    Parameters
    ----------
    score1 : 1-D Tensor
        Results of model A of shape (N, ).
    score2 : 1-D Tensor
        Results of model B of shape (N, ).
    n1 : int
        The number of data points used for training.
    n2 : int
        The number of data points used for testing.
    
    Returns
    -------
    t : float
        The t-statistic.
    pvalue : float
        Two-tailed p-value. If the chosen significance level is larger than the
        p-value, we reject the null hypothesis and accept that there are signi-
        ficant differences in the two compared models.

    Notes
    -----
    heavy development
    '''
    if score1.dim() != 1:
        raise ValueError('The input tensor dimension should be 1, but got '
                         f'{score1.dim()} of score1.')
    if score2.dim() != 1:
        raise ValueError('The input tensor dimension should be 1, but got '
                         f'{score2.dim()} of score2.')

    diff = score1 - score2
    dbar = torch.mean(diff)
    sigma2 = torch.var(diff)
    sigma2Mod = sigma2 * (1/(n1+n2) + (n2/n1))

    tStatic = dbar / torch.sqrt(sigma2Mod)
